from absurdia.absurdia_object import AbsurdiaObject, AbsurdiaObjectsList
from absurdia.api_error import APIError
from absurdia.api_response import APIResponse
from absurdia.resources import ResourceRequestor

class AccountsList(AbsurdiaObjectsList):
    def __init__(self, response: APIResponse):
        super().__init__(objects=response.json["data"], response=response)
        
class Account(AbsurdiaObject):
    def __init__(self, response: APIResponse):
        super().__init__(response=response)

class AccountsRequestor(ResourceRequestor):
    @property
    def base_path(self):
        return "/v1/accounts"
    
    def from_response(self, response: APIResponse, is_list: bool = False):
        if not response.ok:
            raise APIError(response.text, response.status_code, response.headers)
        if is_list:
            return AccountsList(response=response)
        else:
            return Account(response)
    
    def current(self) -> Account:
        path = "%s?current=true" % (self.base_path,)
        response = self._client.request("GET", path)
        return self.from_response(response)


import subprocess
import absurdia
import click

from click import secho, echo
from absurdia.cli.common import check_login
from .freqtrade import read_results, transform_cmd

def upload_freqtrade_results(exportpath: str, 
                             configpath: str, 
                             name: str = None, 
                             cli_command: str = None):
    """Uploads Freqtrade results to Absurdia"""

    data = read_results(exportpath, configpath)
    client = absurdia.Client(agent=absurdia.token)
    
    try:
        backtest = client.backtests.import_freqtrade(data, name, cli_command)
        secho("Successfully imported!", fg='green')
        bid = backtest["id"]
        sid = backtest["strategy_id"]
        # Show page where the backtest is available
        echo("""The results are available at: 
             https://app.absurdia.markets/backtesting/strategies/{}/backtests/{}
             """.format(sid, bid)
        )
    except Exception as e:
        secho("Failed to upload backtest. %s" % (e,), fg="red", err=True)

@click.command(name="import", context_settings={"ignore_unknown_options": True})
@click.option('-n', '--name', type=str, 
              help="A name for the backtest. If not given, will be created randomly.")
@click.option('--freqtrade', nargs=0, 
              help="""A `freqtrade backtesting` command to run.
              The result of the backtest will be uploaded to Absurdia.""")
@click.option('-a', '--adapter', type=str, 
              help="Adapter to use.")
@click.option('-p', '--params', type=click.Path(exists=True, dir_okay=False), 
              help="Path to parameters/configs file (JSON).")
@click.option('-d', '--data', type=click.Path(exists=True, dir_okay=False), 
              help="Path to data file (JSON).")
@click.argument('command', nargs=-1, required=False)
def _import(name, freqtrade, adapter, params, data, command):
    """
    Backtesting service to automatically upload the results of
    a backtest once it has finished.
    Example: absurdia backtest --freqtrade backtesting
    """
    check_login()

    if isinstance(freqtrade, tuple) and command:
        cmd = ("freqtrade",) + command
        secho("[Absurdia] Running Freqtrade command: " + " ".join(cmd), fg='blue')
        fin = transform_cmd(cmd)
        process = subprocess.Popen(fin["command"])
        process.wait()
        upload_freqtrade_results(
            fin["exportpath"], 
            fin["configpath"], 
            name=name, 
            cli_command=" ".join(cmd)
        )
    elif adapter == 'freqtrade':
        if not data:
            secho(
                "Invalid import command. Missing a `--data` argument.", 
                fg='red', 
                err=True
            )
            return
        else:
            upload_freqtrade_results(data, params)
    else:
        echo(click.style("Invalid import command.", fg='red'), err=True)


import os
import time
import json

def transform_cmd(cmd: tuple) -> dict:
    """Returns a dictionary with new command and metadata to import.
    return {
        command: dict,
        configpath: string (file path) or None,
        exportpath: string (file path)
    }
    """
    new_cmd = list(cmd)
    result = {
        "exportpath": None,
        "configpath": None,
    }
    for i, e in enumerate(cmd):
        if e == "-c" or e.startswith("--config"):
            if len(e.split("=")) > 1:
                result["configpath"] = e.split("=")[1]
            else:
                result["configpath"] = cmd[i+1]
        if e.startswith("--export-filename"):
            if len(e.split("=")) > 1:
                result["exportpath"] = e.split("=")[1]
            else:
                result["exportpath"] = cmd[i+1]
        if e.startswith("--export"): # Force export of trades if set to anything else
            if len(e.split("=")) > 1:
                new_cmd[i+1] = "trades"

    if result["exportpath"] is None: # No export path specified
        result["exportpath"] = "backtest_{}".format(str(round(time.time())))
        new_cmd.append("--export-filename={}.json".format(result["exportpath"]))
        
    if result["configpath"] is None: # No config path specified
        result["configpath"] = "./user_data/config.json"
    
    result["command"] = new_cmd
    return result

def bundle_results(exportpath: str, configpath: str) -> dict:
    """Bundle the results into a dictionary"""
    result = { "adapter": "freqtrade" }

    import freqtrade
    result["framework"] = {
        "name" : "freqtrade",
        "version": freqtrade.__version__
    }
    
    for file in os.listdir("."):
        if file.startswith(exportpath):
            if file.endswith(".meta.json"):
                continue
            elif file.endswith(".json"):
                with open(file, "r") as f:
                    exported = json.load(f)
                    result["data"] = exported
    
    return result

def read_results(exportpath: str, configpath: str) -> dict:
    result = None
    for file in os.listdir("."):
        if file.startswith(exportpath):
            if file.endswith(".meta.json"):
                continue
            elif file.endswith(".json"):
                with open(file, "r") as f:
                    exported = json.load(f)
                    result = exported
    return result


import os
import absurdia
import click

from click import secho

from .importers import _import

os.environ['ABSURDIA_BACKTEST'] = '1'

def validate_token(ctx, param, value):
    if value:
        if len(value) != 64:
            raise click.BadParameter("Invalid agent token.")
    return value

@click.group()
def cli():
    return

@cli.command()
@click.option('--save/--no-save', default=True, help="Print verbose log messages.")
@click.option('-f','--file', type=click.File('rb'), help="""
              Agent credentials file (downloaded from your dashboard).""")
@click.option('-t','--token', type=str, callback=validate_token, help="Agent token.")
@click.option('-s','--sig-key', type=str, help="Agent signature key.")
def login(save, file, token, sig_key):
    """
    Set an API credentials of one of your agent to interact with Absurdia.
    Example: absurdia login --token abcdefghijklmnopqrstuvwxyz0123456789
            --sig-key abcdefghijklmnopqrstuvwxyz0123456789
    """
    
    if not (file or token):
        secho("One of --file or --token must be specified.", fg='red')
        return
    
    if token:
        os.environ['ABSURDIA_TOKEN'] = token
        absurdia.token = token
        if sig_key:
            os.environ['ABSURDIA_SIG_KEY'] = sig_key
            absurdia.agent_signature_key = sig_key        
    elif file:
        absurdia.util.load_agent_from_filecontent(file.read())
    
    # Test we can get the agent's data
    client = absurdia.Client(absurdia.token)

    response = client.agents.current()
    if response:
        secho("Successfully logged agent {} in.".format(response['name']), fg='green')
        if save:
            absurdia.util._save_agent()
    else:
        secho(
            "Failed to log in with given credentials. API Response: %s" % (response), 
            fg='red'
        )


cli.add_command(login)
cli.add_command(_import)



from absurdia.api_error import APIError
from absurdia.clients import Client

class ResourceRequestor:
    
    def __init__(self, client: Client):
        self._client = client
    
    def from_response(self, response, is_list=False):
        raise NotImplementedError("Method not implemented.")
    
    def retrieve(self, id: str, params: dict = {}, additional_headers: dict = {}):
        path = "%s/%s" % (self.base_path, id)
        response = self._client.request(
            "GET", path, params=params, additional_headers=additional_headers
        )
        if not response.ok:
            raise APIError(response.text, response.status_code, response.headers)
        return self.from_response(response)
    
    def list(self, params: dict = { "limit": 100 }, additional_headers: dict = {}):
        if params.get("limit"):
            if params["limit"] > 10000:
                raise ValueError("Limit is too large. Its maximum value is 10000.")

        response = self._client.request(
            "GET", self.base_path, params=params, additional_headers=additional_headers
        )
        if not response.ok:
            raise APIError(response.text, response.status_code, response.headers)
        return self.from_response(response, is_list=True)

    def create(self, data: dict = {}, additional_headers: dict = {}, timeout=5000):
        response = self._client.request(
            "POST", self.base_path, data=data, 
            additional_headers=additional_headers, timeout=timeout
        )
        if not response.ok:
            raise APIError(response.text, response.status_code, response.headers)
        return self.from_response(response)
    
    def update(self, 
               id:str, 
               data: dict = {}, 
               additional_headers: dict = {}, 
               timeout=5000):
        path = "%s/%s" % (self.base_path, id)
        response = self._client.request(
            "PATCH", path, data=data, 
            additional_headers=additional_headers, timeout=timeout
        )
        if not response.ok:
            raise APIError(response.text, response.status_code, response.headers)
        return self.from_response(response)

        
    def delete(self, id: str):
        path = "%s/%s" % (self.base_path, id)
        response = self._client.request("DELETE", path)
        if not response.ok:
            raise APIError(response.text, response.status_code, response.headers)


# Absurdia Python bindings

from absurdia.resources import *  # noqa
from absurdia.version import VERSION
from absurdia.clients import Client
from absurdia.agent_credentials import (
    token, 
    agent_filepath, 
    agent_id, 
    agent_signature_key
)

__all__ = [
    Client,
    token,
    agent_filepath,
    agent_id,
    agent_signature_key
]

# Configuration variables

default_http_client = None
app_info = None
enable_telemetry = True
max_network_retries = 5

# Set to either 'WARNING', 'INFO', 'DEBUG'
log = 'WARNING'

__version__ = VERSION

# Takes a name and optional version and plugin URL.
def set_app_info(name, partner_id=None, url=None, version=None):
    global app_info
    app_info = {
        "name": name,
        "partner_id": partner_id,
        "url": url,
        "version": version,
    }


import json

class APIResponse():
    def __init__(self, status_code: int, text: str, headers=None):
        self.content = text
        self.headers = headers
        self.cached = False
        self.status_code = status_code
        self.ok = self.status_code < 400

    @property
    def text(self):
        return self.content
    
    @property
    def json(self):
        return json.loads(self.content)
    
    @property
    def request_id(self):
        try:
            return self.headers["x-request-id"]
        except KeyError:
            return None
    
    @property
    def idempotency_key(self):
        try:
            return self.headers["idempotency-key"]
        except KeyError:
            return None

    def __repr__(self):
        return 'HTTP {} {} {}'.format(self.status_code, self.content, self.request_id)


import json

class APIError(Exception):
    def __init__(
        self,
        message=None,
        http_status=None,
        headers=None
    ):
        super().__init__(message)

        if message and hasattr(message, "decode"):
            try:
                message = message.decode("utf-8")
            except BaseException:
                message = (
                    "<Could not decode message as utf-8. "
                    "Please report to support@absurdia.markets>"
                )

        self._message = message
        self.http_status = http_status
        self.headers = headers or {}
        self.request_id = self.headers.get("x-request-id", None)
        
        self.json_body = None
        if self.headers.get("content-type", None):
            if "application/json" in self.headers["content-type"]:
                try:
                    self.json_body = json.loads(self._message)
                except BaseException:
                    pass

        self.code = None
        if self.json_body:
            self.code = self.json_body.get("code", None)

    def __str__(self):
        if self.request_id is not None:
            return u"Request {0}: {1}".format(self.request_id, self._message or "<empty>")
        else:
            return self._message

    # Returns the underlying `Exception` (base class) message, which is usually
    # the raw message returned by Absurdia's API.
    @property
    def message(self):
        return self._message
    
    @property
    def json(self):
        return self.json_body
    
    @property
    def status_code(self):
        return self.http_status
    
    @property
    def code(self):
        return self.code

    def __repr__(self):
        return "%s(message=%r, http_status=%r, request_id=%r)" % (
            self.__class__.__name__,
            self.message,
            self.status_code,
            self.request_id,
        )
        
class AuthenticationError(APIError):
    pass


import io
import logging
import sys
import os
import psutil
import platform
import datetime
import calendar
import time, json, base64
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from pathlib import Path
from collections import OrderedDict

try:
    pd = __import__("pandas")
except ImportError:
    pd = None

import absurdia

dirname = os.path.dirname(__file__)
homedir = str(Path.home())
saved_agent_path = homedir + "/.absurdia/agent.env"

ABSURDIA_LOG = os.environ.get("ABSURDIA_LOG", default="warn")

logger = logging.getLogger("absurdia")

__all__ = [
    "io",
    "utf8",
    "log_info",
    "log_debug",
    "logfmt",
]

def current_timestamp(granularity="us"):
    if granularity == "us":
        return int(datetime.datetime.now().timestamp() * 1e6)
    elif granularity == "ms":
        return int(datetime.datetime.now().timestamp() * 1e3)
    elif granularity == "s":
        return int(datetime.datetime.now().timestamp())
    else:
        raise ValueError("Invalid granularity: {}".format(granularity))

def now_ms():
    return int(round(time.time() * 1000))


def to_df(obj):
    if pd is None:
        raise ImportError(
            "`pandas` is required to convert to a DataFrame. "
            "Install with `pip install pandas`"
        )
    else:
        return pd.DataFrame(obj, columns=obj[0].keys())

def utf8(value):
    return value.encode("utf-8")

def is_appengine_dev():
    return "APPENGINE_RUNTIME" in os.environ and "Dev" in os.environ.get(
        "SERVER_SOFTWARE", ""
    )

def _console_log_level():
    if absurdia.log in ["debug", "info"]:
        return absurdia.log
    elif ABSURDIA_LOG in ["debug", "info"]:
        return ABSURDIA_LOG
    else:
        return None


def log_debug(message, **params):
    msg = logfmt(dict(message=message, **params))
    if _console_log_level() == "debug":
        print(msg, file=sys.stderr)
    logger.debug(msg)


def log_info(message, **params):
    msg = logfmt(dict(message=message, **params))
    if _console_log_level() in ["debug", "info"]:
        print(msg, file=sys.stderr)
    logger.info(msg)

def logfmt(props):
    def fmt(key, val):
        return u"{key}={val}".format(key=str(key), val=str(val))

    return u" ".join([fmt(key, val) for key, val in sorted(props.items())])

# Load agent credentials
def load_agent():
    if os.environ.get('ABSURDIA_TOKEN') is not None:
        absurdia.token = os.environ['ABSURDIA_TOKEN']
    if os.environ.get('ABSURDIA_SIG_KEY') is not None:
        absurdia.agent_signature_key = os.environ['ABSURDIA_SIG_KEY']
    if os.environ.get('ABSURDIA_AGENT_ID') is not None:
        absurdia.agent_id = os.environ['ABSURDIA_AGENT_ID']
    if os.path.exists(absurdia.agent_filepath) \
    and (absurdia.token is None \
    or absurdia.agent_signature_key is None \
    or absurdia.agent_id is None):
        file = open(absurdia.agent_filepath, "r").read()
        load_agent_from_filecontent(file)
    if os.path.exists(saved_agent_path) \
    and (absurdia.token is None \
    or absurdia.agent_signature_key is None \
    or absurdia.agent_id is None):
        file = open(saved_agent_path, "r").read()
        load_agent_from_filecontent(file)

def _save_agent():
    if absurdia.token is not None:
        home_agent_file = open(saved_agent_path, "w+")
        home_agent_file.write('\n'.join([
            "ABSURDIA_TOKEN=" + absurdia.token,
            "ABSURDIA_SIG_KEY=" + absurdia.agent_signature_key or "",
            "ABUSRDIA_AGENT_ID=" + absurdia.agent_id or ""
        ]))
        home_agent_file.close()

def load_agent_from_filecontent(content: str, save=False):
    idx = content.find("ABSURDIA_TOKEN")
    if idx > -1 and absurdia.token is None:
        idx = idx + (len("ABSURDIA_TOKEN") + 1)
        absurdia.token = content[idx:content.find("\n", idx)]
    idx = content.find("ABSURDIA_SIG_KEY")
    if idx > -1 and absurdia.agent_signature_key is None:
        idx = idx + (len("ABSURDIA_SIG_KEY") + 1)
        absurdia.agent_signature_key = content[idx:content.find("\n", idx)]
    idx = content.find("ABSURDIA_AGENT_ID")
    if idx > -1 and absurdia.agent_id is None:
        idx = idx + (len("ABSURDIA_AGENT_ID") + 1)
        absurdia.agent_id = content[idx:content.find("\n", idx)]
    if save and absurdia.token:
        _save_agent()

def _encode_datetime(dttime):
    if dttime.tzinfo and dttime.tzinfo.utcoffset(dttime) is not None:
        utc_timestamp = calendar.timegm(dttime.utctimetuple())
    else:
        utc_timestamp = time.mktime(dttime.timetuple())

    return int(utc_timestamp)

def _encode_nested_dict(key, data, fmt="%s[%s]"):
    d = OrderedDict()
    for subkey, subvalue in data.items():
        d[fmt % (key, subkey)] = subvalue
    return d

def _api_encode(data):
    for key, value in data.items():
        key = utf8(key)
        if value is None:
            continue
        elif isinstance(value, list) or isinstance(value, tuple):
            for i, sv in enumerate(value):
                if isinstance(sv, dict):
                    subdict = _encode_nested_dict("%s[%d]" % (key, i), sv)
                    for k, v in _api_encode(subdict):
                        yield (k, v)
                else:
                    yield ("%s[%d]" % (key, i), utf8(sv))
        elif isinstance(value, dict):
            subdict = _encode_nested_dict(key, value)
            for subkey, subvalue in _api_encode(subdict):
                yield (subkey, subvalue)
        elif isinstance(value, datetime.datetime):
            yield (key, _encode_datetime(value))
        else:
            yield (key, utf8(value))

def dump(payload):
    return (json.dumps(payload, separators=(',', ':')) if len(payload) else '')

def sign(payload: dict = {}) -> str:
    global PRIVATE_KEY
    if PRIVATE_KEY is None:
        if absurdia.agent_signature_key:
            PRIVATE_KEY = Ed25519PrivateKey.from_private_bytes(
                base64.b64decode(absurdia.agent_signature_key)
            )
        else:
            raise RuntimeError(
                "No signature key provided. "
                "A signature key is required for endpoints that require signatures."
                )
    timestamp = int(time.time_ns() / 1000)
    payload_bytes = f"{timestamp}.{dump(payload)}".encode()
    print(payload_bytes)
    sig_raw = PRIVATE_KEY.sign(payload_bytes)
    signature = base64.urlsafe_b64encode(sig_raw).replace(b'=', b'').decode()
    return f"t={timestamp},s={signature}"

def convert_to_dict(obj):
    """Converts a AbsurdiaObject back to a regular dict.
    Nested AbsurdiaObject are also converted back to regular dicts.
    :param obj: The AbsurdiaObject to convert.
    :returns: The AbsurdiaObject as a dict.
    """
    if isinstance(obj, list):
        return [convert_to_dict(i) for i in obj]
    # This works by virtue of the fact that AbsurdiaObject _are_ dicts. The dict
    # comprehension returns a regular dict and recursively applies the
    # conversion to each value.
    elif isinstance(obj, dict):
        return {k: convert_to_dict(v) for k, v in iter(obj)}
    else:
        return obj

def get_host_info() -> dict:
    return {
        "absurdia_pkg_version": absurdia.__version__,
        "absurdia_pkg_lang": "python",
        "python_version":  sys.version,
        "hostname": platform.node(),
        "os": platform.system(),
        "arch": platform.machine(),
        "platform_version": platform.version(),
        "cpu": {
            "physical_cores": psutil.cpu_count(logical=False),
            "logical_cores": psutil.cpu_count(logical=True),
            "max_freq_mhz": psutil.cpu_freq().max
        },
        "memory_total": psutil.virtual_memory().total
    }


import absurdia
from click import echo, secho

def check_login():
    absurdia.util.load_agent()
    if absurdia.token is None:
        secho("No agent is logged in.", fg="red", err=True)
        echo("Use the command `absurdia login` to log an agent in.")


from absurdia.absurdia_object import AbsurdiaObject
from absurdia.api_error import APIError
from absurdia.api_response import APIResponse
from absurdia.resources import ResourceRequestor


class User(AbsurdiaObject):
    def __init__(self, response: APIResponse):
        super().__init__(response=response)

class UsersRequestor(ResourceRequestor):
    
    @property
    def base_path(self):
        return "/v1/users"
    
    def from_response(self, response: APIResponse, is_list: bool = False):
        if not response.ok:
            raise APIError(response.text, response.status_code, response.headers)
        
        if is_list:
            return NotImplementedError("Users lists are not supported yet.")
        else:
            return User(response)
        
    def list(self, params: dict = { "limit": 100 }, additional_headers: dict = {}):
        return NotImplementedError("This method is not allowed for the resource Users.")
    
    def current(self) -> User:
        path = "%s" % (self.base_path,)
        response = self._client.request("GET", path)
        return self.from_response(response)



import platform
import json
from absurdia.clients.http_client import HttpClient
from absurdia.util import load_agent
from absurdia.api_error import AuthenticationError
from absurdia.version import VERSION
from absurdia.agent_credentials import token

class Client():
    """ A client for accessing the Absurdia API. """
    
    def __init__(self, 
                 agent: str = None, 
                 test: bool = False, 
                 enable_telemetry: bool = True, 
                 log_level='WARNING'):

        if agent is None:
            load_agent()
            if token is None:
                raise AuthenticationError(
                    "No agent token provided. You can generate agent keys "
                    "from the Absurdia web interface. "
                    "See https://app.absurdia.markets/agents"
                )
            agent = token
        if len(agent) != 64:
            raise AuthenticationError(
                "Bad agent token provided. You can generate agent keys "
                "from the Absurdia web interface. "
                "See https://app.absurdia.markets/agents"
            )
        self.agent = agent
        
        ua = {
            "bindings_version": VERSION,
            "lang": "python",
            "publisher": "absurdia"
        }
        for attr, func in [
            ["lang_version", platform.python_version],
            ["platform", platform.platform],
            ["uname", lambda: " ".join(platform.uname())],
        ]:
            try:
                val = func()
            except Exception:
                val = "(disabled)"
            ua[attr] = val

        self.base_headers = {
            "Abs-Client-Agent": json.dumps(ua),
            "User-Agent": "Absurdia/v1 PythonBindings/%s" % (VERSION,),
            "Authorization": "Bearer %s" % (self.agent,),
            "Content-Type": "application/json",
            "Accept-Encoding": "gzip"
        }
        
        self._test = test
        self.enable_telemetry = enable_telemetry

        self.http_client = HttpClient(log_level=log_level)
        self.__last_response = None
        
        # ResourceRequestors
        self._accounts = None
        self._users = None
        self._agents = None
        self._strategies = None
        self._backtests = None
                
    @property
    def headers(self):
        if self.enable_telemetry and self.__last_response:
            if "x-request-id" in self.__last_response.headers:
                request_id = self.__last_response.headers["x-request-id"]
                telemetry = {
                    "request_metrics": {
                        "request_id": request_id,
                        "duration_ms": self.http_client.last_request_duration_ms or 0
                    }
                }
                self.base_headers["Abs-Client-Telemetry"] = json.dumps(telemetry)
        return self.base_headers
    
    @property
    def hostname(self):
        if self._test:
            return "https://test.api.absurdia.markets"
        return "https://api.absurdia.markets"
    
    def request(self, 
                method: str, 
                path: str, 
                params: dict = {}, 
                data: dict = {}, 
                additional_headers: dict = {},
                timeout: int = 5000
    ):
        """
        Makes a request to the Absurdia API using the configured http client
        Authentication information is automatically added.
        :param str method: HTTP Method
        :param str uri: Fully qualified url
        :param dict[str, str] params: Query string parameters
        :param dict[str, str] data: Request body data
        :param dict[str, str] additional_headers: HTTP Headers
        :param int timeout: Timeout in milliseconds
        :returns: Response from the API
        :rtype: absurdia.api_response.APIResponse
        """
        if not bool(params):
            params = None
        if not bool(data):
            data = None
        
        headers = self.headers.copy()
        headers.update(additional_headers)
        
        response = self.http_client.request(
            method,
            "%s%s" % (self.hostname, path),
            params=params,
            data=data,
            headers=headers,
            timeout=timeout
        )
        self.__last_response = response
        return response
        
    def __repr__(self):
        """
        Provide a friendly representation
        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Absurdia Client - Agent {}>'.format(self.agent)
    
    @property
    def accounts(self):
        if self._accounts is None:
            from absurdia.resources.accounts import AccountsRequestor
            self._accounts = AccountsRequestor(self)
        return self._accounts
    
    @property
    def users(self):
        if self._users is None:
            from absurdia.resources.users import UsersRequestor
            self._users = UsersRequestor(self)
        return self._users
    
    @property
    def agents(self):
        if self._agents is None:
            from absurdia.resources.agents import AgentsRequestor
            self._agents = AgentsRequestor(self)
        return self._agents
    
    @property
    def strategies(self):
        if self._strategies is None:
            from absurdia.resources.strategies import StrategiesRequestor
            self._strategies = StrategiesRequestor(self)
        return self._strategies
    
    @property
    def backtests(self):
        if self._backtests is None:
            from absurdia.resources.backtests import BacktestsRequestor
            self._backtests = BacktestsRequestor(self)
        return self._backtests


import os
from codecs import open
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

os.chdir(here)

with open(
    os.path.join(here, "README.md"), "r", encoding="utf-8"
) as fp:
    long_description = fp.read()

version_contents = {}
with open(os.path.join(here, "absurdia", "version.py"), encoding="utf-8") as f:
    exec(f.read(), version_contents)

setup(
    name="absurdia",
    version=version_contents["VERSION"],
    description="Python bindings for the Absurdia API",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Absurdia",
    author_email="support@absurdia.markets",
    url="https://github.com/absurdia/absurdia-py",
    license="MIT",
    keywords="absurdia api trading",
    packages=find_packages(exclude=["tests", "tests.*"]),
    zip_safe=True,
    install_requires=[
        'click',
        'brotlipy',
        'psutil',
        "colorama",
        "requests"
    ],
    python_requires=">=3.4, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    entry_points={
        "console_scripts": [
            "absurdia = absurdia.cli:cli"
        ]
    },
    project_urls={
        "Bug Tracker": "https://github.com/absurdia/absurdia-py/issues",
        "Changes": "https://github.com/absurdia/absurdia-py/blob/master/CHANGELOG.md",
        "Source Code": "https://github.com/absurdia/absurdia-py",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    setup_requires=["wheel"],
)


import logging
from requests import Request, Session, hooks
from requests.adapters import HTTPAdapter
from urllib.parse import urlencode

from absurdia.api_response import APIResponse
from absurdia.util import now_ms

_logger = logging.getLogger('absurdia.http_client')


class HttpClient():
    """
    General purpose HTTP Client for interacting with the Absurdia API
    """

    def __init__(self, 
                 pool_connections=True, 
                 request_hooks=None, 
                 timeout=None, 
                 logger=_logger, 
                 log_level='WARNING', 
                 proxy=None,
                 max_retries=None):
        """
        Constructor for the HttpClient
        :param bool pool_connections
        :param request_hooks
        :param int timeout: Timeout for the requests, in milliseconds.
                            Timeout should never be zero (0) or less.
        :param logger
        :param dict proxy: Http proxy for the requests session
        :param int max_retries: Maximum number of retries each request should attempt
        """
        self.session = Session() if pool_connections else None
        if self.session and max_retries is not None:
            self.session.mount('https://', HTTPAdapter(max_retries=max_retries))
        self.last_request = None
        self.last_response = None
        self.logger = logger
        self.logger.setLevel(log_level)
        self.request_hooks = request_hooks or hooks.default_hooks()

        if timeout is not None and timeout <= 0:
            raise ValueError("Timeout should never be zero (0) or less.")
        self.timeout = timeout
        self.proxy = proxy if proxy else {}
        
        self.last_request_duration_ms = 0

    def request(self, method, url, params=None, data=None, headers=None, timeout=None,
                allow_redirects=False):
        """
        Make an HTTP Request with parameters provided.
        :param str method: The HTTP method to use
        :param str url: The URL to request
        :param dict params: Query parameters to append to the URL
        :param dict data: Dict to go as JSON in the body of the HTTP request
        :param dict headers: HTTP Headers to send with the request
        :param tuple auth: Basic Auth arguments
        :param float timeout: Socket/Read timeout for the request
        :param boolean allow_redirects: Whether or not to allow redirects
        See the requests documentation for explanation of all these parameters
        :return: An http response
        :rtype: A :class:`APIResponse <absurdia.api_response.APIResponse>` object
        """
        if timeout is not None and timeout <= 0:
            raise ValueError(timeout)

        kwargs = {
            'method': method.upper(),
            'url': url,
            'params': params,
            'json': data,
            'headers': headers,
            'hooks': self.request_hooks
        }

        self._log_request(kwargs)

        self.last_response = None
        session = self.session or Session()
        request = Request(**kwargs)
        self.last_request = request

        prepped_request = session.prepare_request(request)

        settings = session.merge_environment_settings(
            prepped_request.url, self.proxy, None, None, None)

        settings['allow_redirects'] = allow_redirects
        settings['timeout'] = timeout if timeout is not None else self.timeout

        request_start_time = now_ms()
        response = session.send(prepped_request, **settings)
        self.last_request_duration_ms = request_start_time - now_ms()
        
        self._log_response(response)

        self.last_response = APIResponse(
            int(response.status_code), response.text, response.headers)

        return self.last_response

    def _log_request(self, kwargs):
        self.logger.info('-- BEGIN Absurdia API Request --')

        if kwargs['params']:
            self.logger.info('{} Request: {}?{}'.format(
                kwargs['method'], kwargs['url'], urlencode(kwargs['params']))
            )
            self.logger.info('Query Params: {}'.format(kwargs['params']))
        else:
            self.logger.info('{} Request: {}'.format(kwargs['method'], kwargs['url']))

        if kwargs['headers']:
            self.logger.info('Headers:')
            for key, value in kwargs['headers'].items():
                # Do not log authorization headers
                if 'authorization' not in key.lower():
                    self.logger.info('{} : {}'.format(key, value))
        self.logger.debug("Request Body: {}".format(kwargs['json']))
        self.logger.info('-- END Absurdia API Request --')

    def _log_response(self, response):
        self.logger.info('Response Status Code: {}'.format(response.status_code))
        self.logger.info('Response Headers: {}'.format(response.headers))
        self.logger.debug('Reponse Body: {}'.format(response.text))


from logging import warning
from absurdia.absurdia_object import AbsurdiaObject, AbsurdiaObjectsList
from absurdia.api_error import APIError
from absurdia.api_response import APIResponse
from absurdia.resources import ResourceRequestor
from absurdia.util import get_host_info

class BacktestsRequestor(ResourceRequestor): 
    @property
    def base_path(self):
        return "/v1/backtests"
    
    def from_response(self, response: APIResponse, is_list: bool = False):
        if not response.ok:
            raise APIError(response.text, response.status_code, response.headers)
        if is_list:
            return BacktestsList(response)
        else:
            return Backtest(response)
    
    def create(
            self, 
            strategy_id: str,
            start_date: int,
            end_date: int,
            timeframe: str,
            name: str = None,
            description: str = None, 
            metadata: dict = None,
            is_public: bool = False,
            initial_balance: float = 1000.0,
            quote_currency: str = "USD",
            symbols: list = None,
            venues: list = None,
            stop_loss_ratio: float = None,
            epsilon: float = 0.01,
            markets_change: float = None,
            configs: dict = None,
            framework_name: str = None,
            framework_version: str = None
        ):
        data = { 
            "strategy_id": strategy_id,
            "start_date": start_date,
            "end_date": end_date,
            "timeframe": timeframe,
            "initial_balance": initial_balance,
            "is_public": is_public,
            "host": get_host_info()
        }
        if name: data["name"] = name
        if description: data["description"] = description
        if metadata: data["metadata"] = metadata
        if quote_currency: data["quote_currency"] = quote_currency
        if symbols: data["symbols"] = symbols
        if venues: data["venues"] = venues
        if stop_loss_ratio: data["stop_loss_ratio"] = stop_loss_ratio
        if epsilon: data["epsilon"] = epsilon
        if markets_change: data["markets_change"] = markets_change
        if configs: data["configs"] = configs
        if framework_name: 
            data["framework"] = { 
                "name": framework_name,
                "version": framework_version
            }

        response = self._client.request(
            "POST", self.base_path, data=data)
        return self.from_response(response)
    
    def import_freqtrade(
        self, 
        result: any,
        name: str = None,
        cli_command: str = None,
        host: dict = None, 
    ):
        data = {
            "adapter": "freqtrade",
            "data": result
        }
        try:
            from freqtrade import __version__
            data["framework"] = {
                "name" : "freqtrade",
                "version": __version__
            }
        except ImportError:
            warning(
                "Cannot import `freqtrade`. "
                "The version of the library will be set at 0.0.0."
            )
            data["framework"] = {
                "name": "freqtrade",
                "version": "0.0.0"
            }
        
        if host:
            data["host"] = host
        else:
            data["host"] = get_host_info()
        if name:
            data["name"] = name
        if cli_command:
            data["cli_command"] = cli_command
        
        path = "%s/import" % (self.base_path,)
        response = self._client.request(
            "POST", path, data=data, timeout=60000
        )
        return self.from_response(response)
    
class BacktestsList(AbsurdiaObjectsList):
    def __init__(self, response: APIResponse):
        super().__init__(objects=response.json["data"], response=response)
        
class Backtest(AbsurdiaObject):
    def __init__(self, response: APIResponse):
        super().__init__(response=response)
        
    def add_positions(self, positions: list):
        data = { "positions": positions }
        url = "%s/%s/positions" % (self.base_path, self.id)
        response = self._client.request("POST", url, data=data)
        if not response.ok:
            raise APIError(response.text, response.status_code, response.headers)
    
    def add_position(self, position: dict):
        return self.add_positions([position])
    
    def finish(self):
        url = "%s/%s" % (self.base_path, self.id)
        data = { "status": "finished" }
        response = self._client.request("PATCH", url, data=data)
        if not response.ok:
            raise APIError(response.text, response.status_code, response.headers)
        self.__init__(self, response)
    


from absurdia.absurdia_object import AbsurdiaObject, AbsurdiaObjectsList
from absurdia.api_error import APIError
from absurdia.api_response import APIResponse
from absurdia.resources import ResourceRequestor

class StrategiesRequestor(ResourceRequestor): 
    @property
    def base_path(self):
        return "/v1/strategies"
    
    def from_response(self, response: APIResponse, is_list: bool = False):
        if not response.ok:
            raise APIError(response.text, response.status_code, response.headers)
        if is_list:
            return StrategiesList(response)
        else:
            return Strategy(response)
        
    def create(self, name: str, description: str = None, metadata: dict = None):
        data = { "name": name }
        if description: data["description"] = description
        if metadata: data["metadata"] = metadata

        response = self._client.request("POST", self.base_path, data=data)
        return self.from_response(response)
        
class StrategiesList(AbsurdiaObjectsList):
    def __init__(self, response: APIResponse):
        super().__init__(objects=response.json["data"], response=response)
        
class Strategy(AbsurdiaObject):
    def __init__(self, response: APIResponse):
        super().__init__(response=response)
    